/*
 *  By Per Jessen <per@computer.org> with changes by the ClamAV team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 */

#if HAVE_CONFIG_H
#include "clamav-config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <string.h>
#include <errno.h>

#include "shared/output.h"
#include "shared/optparser.h"
#include "execute.h"
/* NPG */
#include "shared/misc.h"
#include <pthread.h>
#include <sys/types.h>
#include <sys/wait.h>
/* NPG End */

#define MAX_CHILDREN 5

int active_children;

void execute( const char *type, const char *text, const struct optstruct *opts )
{
	int ret;

    if(!optget(opts, "daemon")->enabled) {
	if(sscanf(text, "EXIT_%d", &ret) == 1) {
	    logg("*%s: EXIT_%d\n", type, ret);
	    exit(ret);
	}
	if(system(text) == -1)
	    logg("%s: system(%s) failed\n", type, text);

	return;
    }

#ifdef _WIN32
    if(spawnlp(_P_NOWAIT, text, text, NULL) == -1) {
	logg("^%s: couldn't execute \"%s\".\n", type, text);
	return;
    }
#else
/* NPG */
#define CLAM_EVENT "CLAM_EVENT"
#define CLAM_ENGINE "CLAM_ENGINE"
#define CLAM_HOST "CLAM_HOST"
static pthread_mutex_t virusaction_lock = PTHREAD_MUTEX_INITIALIZER;
/* NPG End */
    if ( active_children<MAX_CHILDREN ) {
	pid_t pid;
	/* NPG */
	char *buffer_event, *buffer_engine, *buffer_host;
	char *env[7], host[64];
	const char  *event, *engine;
	size_t i, j, v = 0, len;
	char er[256];
	if ( type == "OnErrorExecute" ) {
		event = "OE";
	}else if ( type == "OnUpdateExecute" ) {
		event = "OU";
	}else if ( type == "OnOutdatedExecute" ) {
		event = "OO";
	}else{
		event = "ER";
	}
	engine = get_version();
	host[63] = '\0';
	gethostname (host, 63);
	env[0] = getenv("PATH");
	j = env[0] ? 1 : 0;
	buffer_event = (char *) malloc(strlen(CLAM_EVENT) + strlen(event) + 2);
	if(buffer_event) {
		sprintf(buffer_event, "%s=%s", CLAM_EVENT, event);
		env[j++] = buffer_event;
	}
	buffer_engine = (char *) malloc(strlen(CLAM_ENGINE) + strlen(engine) + 2);
	if(buffer_engine) {
		sprintf(buffer_engine, "%s=%s", CLAM_ENGINE, engine);
		env[j++] = buffer_engine;
	}
	buffer_host = (char *) malloc(strlen(CLAM_HOST) + strlen(host) + 2);
	if(buffer_host) {
		sprintf(buffer_host, "%s=%s", CLAM_HOST, host);
		env[j++] = buffer_host;
	}
	env[j++] = NULL;
	/* NPG End */
	/*switch( pid=fork() ) {
	case 0:
		if ( -1==system(text) )
		{
		logg("^%s: couldn't execute \"%s\".\n", type, text);
		}
		exit(0);
	case -1:
		logg("^%s::fork() failed, %s.\n", type, strerror(errno));
		break;
	default:
		active_children++;
	}*/
	/* NPG */
	logg("*%s: about to execute '%s' '%s'\n", type, text, event);
	pthread_mutex_lock(&virusaction_lock);
	pid = fork();
	if(!pid) { /* child */
		exit(execle(text, NULL, env));
	} else if(pid > 0) { /* parent */
		active_children++;
		pthread_mutex_unlock(&virusaction_lock);
		/* waitpid(pid, NULL, 0);*/
		if(waitpid(pid, &ret, 0)<0)
			logg("^%s: waitpid() failed: %s\n", type, cli_strerror(errno, er, sizeof(er)));
		else {
			if(WIFEXITED(ret))
				logg("^%s: child exited with code %d\n", type, WEXITSTATUS(ret));
			else if(WIFSIGNALED(ret))
				logg("^%s: child killed by signal %d\n", type, WTERMSIG(ret));
			else
				logg("^%s: child lost\n", type);
		}
	} else {
		logg("^%s::fork() failed, %s.\n", type, strerror(errno));
	}
	free(buffer_event);
	free(buffer_engine);
	free(buffer_host);
	/* NPG End */
    } else {
		logg("^%s: already %d processes active.\n", type, active_children);
    }
#endif
}
